@@ -9,12 +9,12 @@
 */
 
 #include "open-behind-mem-types.h"
-#include "xlator.h"
-#include "statedump.h"
-#include "call-stub.h"
-#include "defaults.h"
+#include <glusterfs/xlator.h>
+#include <glusterfs/statedump.h>
+#include <glusterfs/call-stub.h>
+#include <glusterfs/defaults.h>
 #include "open-behind-messages.h"
-#include "glusterfs-acl.h"
+#include <glusterfs/glusterfs-acl.h>
 
 typedef struct ob_conf {
     gf_boolean_t use_anonymous_fd; /* use anonymous FDs wherever safe
@@ -105,14 +105,14 @@ ob_inode_get(xlator_t *this, inode_t *inode)
             if (ob_inode == NULL)
                 goto unlock;
 
-            value = (uint64_t)((void *)ob_inode);
+            value = (uint64_t)(uintptr_t)ob_inode;
             ret = __inode_ctx_set(inode, this, &value);
             if (ret < 0) {
                 ob_inode_free(ob_inode);
                 ob_inode = NULL;
             }
         } else {
-            ob_inode = (ob_inode_t *)value;
+            ob_inode = (ob_inode_t *)(uintptr_t)value;
         }
     }
 unlock:
@@ -206,8 +206,13 @@ ob_fd_free(ob_fd_t *ob_fd)
     if (ob_fd->xdata)
         dict_unref(ob_fd->xdata);
 
-    if (ob_fd->open_frame)
+    if (ob_fd->open_frame) {
+        /* If we sill have a frame it means that background open has never
+         * been triggered. We need to release the pending reference. */
+        fd_unref(ob_fd->fd);
+
         STACK_DESTROY(ob_fd->open_frame->root);
+    }
 
     GF_FREE(ob_fd);
 }
@@ -297,6 +302,7 @@ ob_wake_cbk(call_frame_t *frame, void *cookie, xlator_t *this, int op_ret,
             call_resume(stub);
     }
 
+    /* The background open is completed. We can release the 'fd' reference. */
     fd_unref(fd);
 
     STACK_DESTROY(frame->root);
@@ -331,7 +337,9 @@ ob_fd_wake(xlator_t *this, fd_t *fd, ob_fd_t *ob_fd)
     }
 
     if (frame) {
-        frame->local = fd_ref(fd);
+        /* We don't need to take a reference here. We already have a reference
+         * while the open is pending. */
+        frame->local = fd;
 
         STACK_WIND(frame, ob_wake_cbk, FIRST_CHILD(this),
                    FIRST_CHILD(this)->fops->open, &ob_fd->loc, ob_fd->flags, fd,
@@ -345,14 +353,13 @@ void
 ob_inode_wake(xlator_t *this, struct list_head *ob_fds)
 {
     ob_fd_t *ob_fd = NULL, *tmp = NULL;
-    fd_t *fd = NULL;
 
-    list_for_each_entry_safe(ob_fd, tmp, ob_fds, ob_fds_on_inode)
-    {
-        ob_fd_wake(this, ob_fd->fd, ob_fd);
-        fd = ob_fd->fd;
-        ob_fd_free(ob_fd);
-        fd_unref(fd);
+    if (!list_empty(ob_fds)) {
+        list_for_each_entry_safe(ob_fd, tmp, ob_fds, ob_fds_on_inode)
+        {
+            ob_fd_wake(this, ob_fd->fd, ob_fd);
+            ob_fd_free(ob_fd);
+        }
     }
 }
 
@@ -363,7 +370,7 @@ ob_fd_copy(ob_fd_t *src, ob_fd_t *dst)
     if (!src || !dst)
         goto out;
 
-    dst->fd = __fd_ref(src->fd);
+    dst->fd = src->fd;
     dst->loc.inode = inode_ref(src->loc.inode);
     gf_uuid_copy(dst->loc.gfid, src->loc.gfid);
     dst->flags = src->flags;
@@ -381,9 +388,7 @@ open_all_pending_fds_and_resume(xlator_t *this, inode_t *inode,
     ob_fd_t *ob_fd = NULL, *tmp = NULL;
     gf_boolean_t was_open_in_progress = _gf_false;
     gf_boolean_t wait_for_open = _gf_false;
-    struct list_head ob_fds = {
-        0,
-    };
+    struct list_head ob_fds;
 
     ob_inode = ob_inode_get(this, inode);
     if (ob_inode == NULL)
@@ -509,7 +514,6 @@ ob_open_behind(call_frame_t *frame, xlator_t *this, loc_t *loc, int flags,
 
     ob_fd->ob_inode = ob_inode;
 
-    /* don't do fd_ref, it'll cause leaks */
     ob_fd->fd = fd;
 
     ob_fd->open_frame = copy_frame(frame);
@@ -539,15 +543,16 @@ ob_open_behind(call_frame_t *frame, xlator_t *this, loc_t *loc, int flags,
     }
     UNLOCK(&fd->inode->lock);
 
-    if (!open_in_progress && !unlinked) {
-        fd_ref(fd);
+    /* We take a reference while the background open is pending or being
+     * processed. If we finally wind the request in the foreground, then
+     * ob_fd_free() will take care of this additional reference. */
+    fd_ref(fd);
 
+    if (!open_in_progress && !unlinked) {
         STACK_UNWIND_STRICT(open, frame, 0, 0, fd, xdata);
 
         if (!conf->lazy_open)
             ob_fd_wake(this, fd, NULL);
-
-        fd_unref(fd);
     } else {
         ob_fd_free(ob_fd);
         STACK_WIND(frame, default_open_cbk, FIRST_CHILD(this),
@@ -576,7 +581,7 @@ ob_open(call_frame_t *frame, xlator_t *this, loc_t *loc, int flags, fd_t *fd,
 {
     fd_t *old_fd = NULL;
     int ret = -1;
-    int op_errno = 0;
+    int op_errno = ENOMEM;
     call_stub_t *stub = NULL;
 
     old_fd = fd_lookup(fd->inode, 0);
@@ -584,7 +589,6 @@ ob_open(call_frame_t *frame, xlator_t *this, loc_t *loc, int flags, fd_t *fd,
         /* open-behind only when this is the first FD */
         stub = fop_open_stub(frame, default_open_resume, loc, flags, fd, xdata);
         if (!stub) {
-            op_errno = ENOMEM;
             fd_unref(old_fd);
             goto err;
         }
@@ -598,7 +602,6 @@ ob_open(call_frame_t *frame, xlator_t *this, loc_t *loc, int flags, fd_t *fd,
 
     ret = ob_open_behind(frame, this, loc, flags, fd, xdata);
     if (ret) {
-        op_errno = ENOMEM;
         goto err;
     }
 
@@ -869,18 +872,12 @@ int
 ob_finodelk(call_frame_t *frame, xlator_t *this, const char *volume, fd_t *fd,
             int cmd, struct gf_flock *flock, dict_t *xdata)
 {
-    call_stub_t *stub = NULL;
-
-    stub = fop_finodelk_stub(frame, default_finodelk_resume, volume, fd, cmd,
-                             flock, xdata);
-    if (!stub)
-        goto err;
-
-    open_and_resume(this, fd, stub);
-
-    return 0;
-err:
-    STACK_UNWIND_STRICT(finodelk, frame, -1, ENOMEM, 0);
+    call_stub_t *stub = fop_finodelk_stub(frame, default_finodelk_resume,
+                                          volume, fd, cmd, flock, xdata);
+    if (stub)
+        open_and_resume(this, fd, stub);
+    else
+        STACK_UNWIND_STRICT(finodelk, frame, -1, ENOMEM, 0);
 
     return 0;
 }
@@ -890,18 +887,12 @@ ob_fentrylk(call_frame_t *frame, xlator_t *this, const char *volume, fd_t *fd,
             const char *basename, entrylk_cmd cmd, entrylk_type type,
             dict_t *xdata)
 {
-    call_stub_t *stub = NULL;
-
-    stub = fop_fentrylk_stub(frame, default_fentrylk_resume, volume, fd,
-                             basename, cmd, type, xdata);
-    if (!stub)
-        goto err;
-
-    open_and_resume(this, fd, stub);
-
-    return 0;
-err:
-    STACK_UNWIND_STRICT(fentrylk, frame, -1, ENOMEM, 0);
+    call_stub_t *stub = fop_fentrylk_stub(
+        frame, default_fentrylk_resume, volume, fd, basename, cmd, type, xdata);
+    if (stub)
+        open_and_resume(this, fd, stub);
+    else
+        STACK_UNWIND_STRICT(fentrylk, frame, -1, ENOMEM, 0);
 
     return 0;
 }
@@ -910,18 +901,12 @@ int
 ob_fxattrop(call_frame_t *frame, xlator_t *this, fd_t *fd,
             gf_xattrop_flags_t optype, dict_t *xattr, dict_t *xdata)
 {
-    call_stub_t *stub = NULL;
-
-    stub = fop_fxattrop_stub(frame, default_fxattrop_resume, fd, optype, xattr,
-                             xdata);
-    if (!stub)
-        goto err;
-
-    open_and_resume(this, fd, stub);
-
-    return 0;
-err:
-    STACK_UNWIND_STRICT(fxattrop, frame, -1, ENOMEM, 0, 0);
+    call_stub_t *stub = fop_fxattrop_stub(frame, default_fxattrop_resume, fd,
+                                          optype, xattr, xdata);
+    if (stub)
+        open_and_resume(this, fd, stub);
+    else
+        STACK_UNWIND_STRICT(fxattrop, frame, -1, ENOMEM, 0, 0);
 
     return 0;
 }
@@ -1109,7 +1094,7 @@ ob_forget(xlator_t *this, inode_t *inode)
     inode_ctx_del(inode, this, &value);
 
     if (value) {
-        ob_inode = (ob_inode_t *)((void *)value);
+        ob_inode = (ob_inode_t *)(uintptr_t)value;
         ob_inode_free(ob_inode);
     }
 
@@ -1305,6 +1290,14 @@ struct xlator_dumpops dumpops = {
 };
 
 struct volume_options options[] = {
+    {
+        .key = {"open-behind"},
+        .type = GF_OPTION_TYPE_BOOL,
+        .default_value = "off",
+        .description = "enable/disable open-behind",
+        .op_version = {GD_OP_VERSION_6_0},
+        .flags = OPT_FLAG_SETTABLE,
+    },
     {
         .key = {"use-anonymous-fd"},
         .type = GF_OPTION_TYPE_BOOL,
@@ -1349,3 +1342,17 @@ struct volume_options options[] = {
     {.key = {NULL}}
 
 };
+
+xlator_api_t xlator_api = {
+    .init = init,
+    .fini = fini,
+    .reconfigure = reconfigure,
+    .mem_acct_init = mem_acct_init,
+    .op_version = {1}, /* Present from the initial version */
+    .dumpops = &dumpops,
+    .fops = &fops,
+    .cbks = &cbks,
+    .options = options,
+    .identifier = "open-behind",
+    .category = GF_MAINTAINED,
+};
